// MyApp.cc

#include <iostream.h>
#include <complex.h>
#include "ocs.hh"

#include "MyStartup.hh"
#include "ocsOptDataWriter.hh"
#include <time.h>
#include <sys/time.h>
#include <sys/times.h>


/*
Declare all functions in this file 
(except for the function MyApplication) here:
*/

void BackToBackNoiseFreeSystem(StartupExample * MyStartup);

void Soliton(StartupExample * Startup);
void Tyco(StartupExample * Startup);

extern ofstream LogFile;

// #######################################################

// ## Application Wrapper

void MyApplication(int argc, char *argv[]) 
{
  /*
  Construct an instance  MyStartup of the class StartupExample
  See MyStartup.hh for the definition of this class 

  Startup.in is the name of the input parameter file for the class
  StartupExample
  */


  StartupExample MyStartup("Startup.in",argc,argv);


  /*
  Read O'Reilly pp. 110-115 for info on switch statements. 
  switch must take an integer argument, in this case
  MyStartup.GetSimulationType()
  */

  switch(MyStartup.GetSimulationType())
   {
    
    case 1:
      BackToBackNoiseFreeSystem(&MyStartup);
      break;

     case 2:
       Soliton(&MyStartup);
       break;  

     case 3:
       Tyco(&MyStartup);
       break;


    default:
      cerr << "Invalid choice of application in SimType" << endl;
      exit(1);
          
   } // end switch

} // ## end MyApplication

// ########################################################

void BackToBackNoiseFreeSystem(StartupExample * MyStartup)
{

  // Read about strings in O'Reilly page 49

  string Job = MyStartup->GetJobPrefix();

  LogFileSeparator();
  LogFile << "BackToBackSystem" << endl;

  /* 
     The RanNumGen class is used to generate random numbers.
     RNG is an instance of the RanNumGen class, i.e. it is 
     a RanNumGen object. The following line declares RNG to be
     a RanNumGen object and also calls the class constructor
     which initializes the data in the class. The input parameters
     for RNG are in the file RanNumGen.in in the directory
     $(HOME)/ENEE785/expts/BackToBackNoiseFreeSystem
  */

  RanNumGen RNG("RanNumGen.in");

  /* The OptSignal class contains the data and methods we need to
     model the optical signal. 
     Signal is an instance of an OptSignal.

     The OptSignal constructor is called in the following line.
     This constructor reads in input parameters from a file (OptSignal.in).
     In the constructor we pass a reference to a RanNumGen object (RNG),
     which tells Signal the memory location of RNG. The RanNumGen object
     is used if we have a random bit string, or random input polarization
     states in the different channels.

     The OptSignal constuctor initializes the variables in the Signal class.
     In particular, it calls the method OptSignal::RegenerateSignalString()
     which initializes the data arrays used to store the signal. These four
     cplx arrays store X and Y polarization states of the signal in both
     the time and frequency domains. 
  */

  OptSignal Signal("OptSignal.in", &RNG);
  
  // Write the initial optical signal power in time and in frequency to files 

  int OmitDCFlag = 0;

  Signal.WriteFileTimePower(Job + "tInOpt.dat");
  Signal.WriteFileFreqPower(Job + "fInOpt.dat",OmitDCFlag);

  /* 
     The following three lines declare and construct
     the components of a receiver - an optical filter, 
     photodetector, and electric filter. The instance of
     the OptFilter class is called oOptFilter, ("o" for object).

     We need to pass a reference to Signal to both oOptFilter
     and oPhotodetector since both these classes operate on the
     data in the signal class. 

     The Photodetector contains two cplx arrays which store the
     time and frequency data for the received electrical current.

     So we must pass a reference to oPhotodetector to oElecFilter
  */
     
  OptFilter oOptFilter("OptFilter.in",&Signal); 
  Photodetector  oPhotodetector("Photodetector.in",&Signal); 
  ElecFilter oElecFilter("ElecFilter.in",&oPhotodetector);

  /*

   The ElecSignalStat class is used to compute the clock recovery time
   and the statistics of the received electrical current, 
   e.g. the Q-factor and the eye opening. This class is large and quite
   complicated.

   We need to pass a reference to an OptSignal because the ElecSignalStat
   class needs to know the BitString, which is an variable in the OptSignal
   class. We also need a reference to oPhotodetector to access the
   data arrays for the received electrical current.
  */

  ElecSignalStat oElecSignalStat(&Signal,&oPhotodetector);   

  /* 
    At last, we can send the signal through the receiver
    and actually compute the statistics of the received signal.
  */

   oOptFilter.FilterOptSignal();

    Signal.WriteFileTimePower(Job + "tAfterDemuxOpt.dat");
    Signal.WriteFileFreqPower(Job + "fAfterDemuxOpt.dat",OmitDCFlag);

   oPhotodetector.DetectOptSignal();   
   oElecFilter.FilterElecSignal();
   oElecSignalStat.GetSingleStringElecSignalStat();

   /* 
     To look at the results in this application we 
     (1) Write the electrical signal power in time and in frequency  to a file
     (2) Write out the Eye diagram to a file
     (3) Write the back to back eye opening to the screen
   */

   oPhotodetector.WriteFileTimePower(Job + "tReceivedCurrent.dat");   
   oPhotodetector.WriteFileFreqPower(Job + "fReceivedCurrent.dat",OmitDCFlag);

   oElecSignalStat.WriteFileEyeDiagram(Job + "Eye.dat"); 

   cout << "Back to Back Eye Opening [mV] = " 
	<< 1e3*oElecSignalStat.GetMinAmplitudeMargin() << endl;

   LogFile << "THE END" << endl;
   cout << "THE END" << endl;

} // ## end Back2BackNoiseFreeSystem

// ############################################

/*!

Soliton propagates a first-order classical soliton 
through an optical fiber. See FFTandNLSConventions.eps
on the course web page for more details.

*/


void Soliton(StartupExample * Startup)
{

  LogFileSeparator();

  LogFile <<  endl << "Soliton" << endl;

  string InDir = "./";
  string Job = Startup->GetJobPrefix();

  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
  RanNumGen RNGFiber(InDir + "RanNumGenFiber.in");

  OptSignal Signal("Signal.in",&RNGSignal);

  // The OptDataWriter class can be used to output optical data
  // to a file. You just need to set flags in DataWriter.in and
  // add the appropriate calls in the code as below and in the next 
  // function Tyco.

  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,&Signal);

  DataWriter.WriteInitialData();
  Signal.WriteComplexFields(Job+"tComplexIn.dat",0);

  // Constructor for class OptFiberLocalError

  OptFiberLocalError Fiber("Fiber.in",&Signal,&RNGFiber,Job);

  // Because of the way Fiber.in is written it is easier
  // to hard-code the parameters for a soliton:

  // Parameters for a first-order soliton

  Fiber.SetFstOrDispFiber(-0.1*1e-27);  // -0.1 ps^2/km
  Fiber.SetSndOrDispFiber(0*1e-39);     // 0 ps^3/km

  // Nonlinear coefficient in NLS 2.2 W^{-1}km^{-1}:

  Fiber.SetGammaFiber(2.2*1e-3);        
  

  /***

  // Uncomment if you want pure third order dispersion rather than a soliton

  Fiber.SetFstOrDispFiber(0);  // -0.1 ps^2/km
  Fiber.SetSndOrDispFiber(1e-39);     // 0 ps^3/km
  Fiber.SetGammaFiber(0);        // 2.2 W^{-1}km^{-1}

  ***/

   // PropagatedLength keeps track of the propagation distance through
  // the fibers in the system

  double PropagatedLength = 0.0;

   // ZStepNum keeps track of number of steps in the distance variable z 
   // that are taken by fiber propagation algorithm

  int ZStepNum = 0; 

  // Finally we call the method that propagates the signal through the fiber

  Fiber.Propagate(&PropagatedLength,&ZStepNum);

  Signal.WriteComplexFields(Job+"tComplexOut.dat",0);

  DataWriter.WriteEndTransmission(0,0,PropagatedLength);

  cout << "PropagatedLength [km] = " << 1e-3*PropagatedLength << endl;
  cout << "Number of Steps = " << ZStepNum << endl;
  cout << "THE END" << endl;

} // ## end Soliton

// #################################################

/*!

This code models a 10 Gb/s chirped return-to-zero undersea 
system based on DSF transmission
fiber and which used SMF for dispersion compensation.
It is modeled on an undersea system of Tyco.
For details on the system see 

R.-M. Mu, T. Yu, V.S. Grigoryan, and C.R. Menyuk, 
"Dynamics of the Chirped Return-to-Zero Modulation Format," J.
Lightwave Technol. 20, 47-57 (2002). 

This is paper #147 on www.photonics.umbc.edu (Recent Publications)

This code gives you a first look at how to put together a full system
model using OCS.

If you want to model noise from optical amplifiers you will need to 
adapt this code.

*/

void Tyco(StartupExample * Startup)
{

  time_t StartTime,EndTime;
  struct tms my_tms;

  StartTime = time(NULL);
  cout << "Start Time: " << ctime(&StartTime) << endl;

  LogFileSeparator();

  LogFile <<  endl << "Tyco" << endl;

  string InDir = "./";
  string Job = Startup->GetJobPrefix();

  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
  RanNumGen RNGAmps(InDir + "RanNumGenAmps.in");
  RanNumGen RNGFibers(InDir + "RanNumGenFibers.in");

  // Declaration of pointer to an OptSignal object

  OptSignal * Signal;

  // Call to OptSignal constructor

  Signal = new OptSignal("Signal.in",&RNGSignal);

  //  OptFilter TransmitterOptFilter("OptFilterTx.in",Signal);
    // Careful re this when do WDM!!

  //TransmitterOptFilter.FilterOptSignal();

  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,Signal);

  DataWriter.WriteInitialData();
 
   // Fiber constructors

    OptFiberLocalError PreDCF(InDir + "PreDCF.in",Signal,&RNGFibers,Job);
    OptFiberLocalError PostDCF(InDir + "PostDCF.in",Signal,&RNGFibers,Job);
    OptFiberLocalError DSF35(InDir + "DSF35.in",Signal,&RNGFibers,Job);
    OptFiberLocalError DSF45(InDir + "DSF45.in",Signal,&RNGFibers,Job);
    OptFiberLocalError SMF(InDir + "SMF.in",Signal,&RNGFibers,Job);
   
    int NumFibers = 5;

    OptFiberLocalError * FiberPtr[NumFibers];

    FiberPtr[0] = &PreDCF;
    FiberPtr[1] = &PostDCF;
    FiberPtr[2] = &DSF45;
    FiberPtr[3] = &DSF35;
    FiberPtr[4] = &SMF;

    // Amp is the amplifier in the dispersion map
    // PAmp is the amplifier after the PreDCF and PostDCF

    OptAmplifier Amp(InDir + "Amp.in",Signal,&RNGAmps);
    OptAmplifier PAmp(InDir + "PAmp.in",Signal,&RNGAmps);

    int NumAmps = 2;

    OptAmplifier * AmpPtr[NumAmps];

    AmpPtr[0] =  &Amp;
    AmpPtr[1] =  &PAmp;

    // These lines of code enable the user to change parameters
    // in all the Fibers and Amps just by setting a parameter
    // in Startup.in

  if(Startup->GetDebugLevelAmps())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetDebugLevel(Startup->GetDebugLevelAmps());
     }
   
  for(int i=0;i<NumAmps;i++)
    AmpPtr[i]->SetTypeAmplifier(typeAmplifier(Startup->GetTypeAmp()));
     
   if(Startup->GetAmpsNoiseOffFlag())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetTypeAmplifierNoise(NOISE_OFF);
     }

   if(Startup->GetBeta3ZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetSndOrDispFiber(0);
     }

   if(Startup->GetNonLinearityZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetGammaFiber(0);
     }

   if(Startup->GetFiberLossZeroFlag()) 
        for(int ii = 0; ii< NumFibers;ii++)
            FiberPtr[ii]->SetAttenuation(0);

    for(int ii = 0; ii< NumFibers;ii++)
         FiberPtr[ii]->SetRelativeErrorGoal(Startup->GetRelativeErrorGoal());

      for(int ii = 0; ii< NumFibers;ii++)
	{
	  FiberPtr[ii]->SetTypeStepSizes( Startup->GetTypeStepSizes());

          if(Startup->GetTypeStepSizes() == 3)
	    FiberPtr[ii]->SetWalkOffParameter(Startup->GetWalkOffParameter());

           FiberPtr[ii]->SetOutputStepSizesFlag(
                            Startup->GetOutputStepSizesFlag());
	}


       // #### Propagate signal through fiber ####

       double PropagatedLength = 0;
       int ZStepNum = 0;

      PreDCF.Propagate(&PropagatedLength,&ZStepNum);
      
      PAmp.AmplifyOptSignal();

      int kk = 0;

      DataWriter.WriteEndMap(0,kk,PropagatedLength);

      for(kk = 0; kk < Startup->GetNumMapSpans(); kk++) 
       {
        SMF.Propagate(&PropagatedLength,&ZStepNum);
        DSF35.Propagate(&PropagatedLength,&ZStepNum);

        Amp.AmplifyOptSignal();

        DSF45.Propagate(&PropagatedLength,&ZStepNum);
        Amp.AmplifyOptSignal();
        DSF45.Propagate(&PropagatedLength,&ZStepNum);
        Amp.AmplifyOptSignal();
     
      
        DSF35.Propagate(&PropagatedLength,&ZStepNum);
        SMF.Propagate(&PropagatedLength,&ZStepNum);
        Amp.AmplifyOptSignal();

        DataWriter.WriteEndMap(0,kk,PropagatedLength);

        if(kk % 10 == 0 && kk != 0)
	 cout << "Completed " << kk << " periods of dispersion map " << endl;

       }// ## end for kk-loop

    PostDCF.Propagate(&PropagatedLength,&ZStepNum);
    PAmp.AmplifyOptSignal();

    // ############### End of Propagation #################

  DataWriter.WriteEndTransmission(0,kk,PropagatedLength);
 

        cout <<"Total PropagatedLength [km] = "<<PropagatedLength*1e-3<< endl;
        cout <<"Number Steps in Z taken by solver = " << ZStepNum << endl;

	// Next we set up an optical demuxer


  OptFilter* TunableFilter;
  TunableFilter = new OptFilter(InDir + "TunableOptFilter.in",
				Signal->GetTypeSimulation());

  OptDemuxer *oOptDemuxer;
  oOptDemuxer = new OptDemuxer(Signal, TunableFilter);

  // SCSignal is a pointer to an OptSignal which is used to store the data
  // for the single channel signal you get after the optical demuxer

  // Passing the argument 0 to OptDemuxer::ExtractChannel demuxes
  // the central channel, if you are using the option
  // OptSignal::EvenlySpacedFreqsFlag = 1

  OptSignal * SCSignal;
  SCSignal = new OptSignal(oOptDemuxer->ExtractChannel(0));

  Photodetector oPhotodetector("Photodetector.in",SCSignal,&RNGSignal);
  oPhotodetector.DetectOptSignal();
  
  ElecFilter * oElecFilter;
  oElecFilter = new ElecFilter(InDir + "ElecFilter.in",&oPhotodetector);

  ElecSignalStat *oElecSignalStat;
  oElecSignalStat = new  ElecSignalStat(SCSignal,&oPhotodetector);
	
  oElecFilter->FilterElecSignal();     

  oPhotodetector.WriteFileTimePower(Job+"tReceivedCurrent.dat");

  // NOTE: You should always use the NEW method as below.
  // I think this was not done in BackToBackNoiseFreeSystem

  oElecSignalStat->GetSingleStringElecSignalStatConvolutingNEW();
  oElecSignalStat->WriteFileEyeDiagram(Job+"Eye.dat"); 

  cout << "Eye Opening [mV] = " 
       << 1e3*oElecSignalStat->GetAmplitudeMargin() << endl;
  cout << "Q factor = " << oElecSignalStat->GetQ_Factor() << endl;

  EndTime = time(NULL);
  double TotalTime = double(EndTime - StartTime);

  times(&my_tms);

  double CPUtime = (double(my_tms.tms_utime))/(double(sysconf(_SC_CLK_TCK)));

  cout << "Total Run Time [hrs] = " << TotalTime/3600.0 << endl
       << "Total CPU Time [hrs] = " << CPUtime/3600.0 << endl
       << "Total CPU Time [secs] = " << CPUtime << " = " 
       << 100.0*CPUtime/TotalTime << "%" << endl
       << "End Time: " << ctime(&EndTime) << endl;

   cout << "THE END" << endl;
   LogFile << "THE END" << endl;

} // ## end Tyco

// ############################################################






